/*
 * @(#)xtertris.c
 *
 * Copyright 1993 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Driver file for Tertris */

#ifndef WINVER
#include "version.h"
static const char aboutHelp[] = {
"Tertris Version " VERSION "\n"
"Send bugs (reports or fixes) to the author: "
"David Bagley <bagleyd AT verizon.net>\n"
"The latest version is at: "
"https://www.sillycycle.com/altris.html"
};

static const char synopsisHelp[] = {
"[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n"
"[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n"
"[-{foreground|fg} {color}] [-{background|bg} {color}]\n"
"[-[no]sound] [-bumpSound {filename}]\n"
"[-level {starting_level}] [-units {size_of_polyiamond}]\n"
"[-columns {columns}] [-fill {rows}]\n"
"[-[no]fixedFill] [-[no]suddenAppear]\n"
"[-[no]diagonal] [-[no]mixed] [-[no]bonus] [-[no]cw]\n"
"[-[no]plain] [-[no]showNext] [-[no]grid]\n"
"[-userName {string}] [-scoreFile {filename}] [-scores]\n"
"[-version]"
};
#endif

#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA) || defined(WINVER)
static const char descriptionHelp[] = {
"The ever popular falling tetromino game now with triangles.  "
"Enhanced to allow all polyiamond classes that do not have\n"
"pieces with holes.  Mixing polyiamond classes (so as to "
"permit polyiamonds made up by a variable number triangles)\n"
"can also be selected, as well as, 1 set of diagonal "
"polyiamonds that do not have pieces with holes.\n"
"\n"
"To make these games interesting and \"do-able\" the ability "
"to rotate gets turned off if it is easier than the 4\n"
"triangle polyiamond.  An added ability \"reflect\" "
"(right-left -> left-right) allows one to attempt the 5\n"
"polyiamond pieces.\n"
"\n"
"To get high scores, do not use \"show next\" mode, try "
"starting at a higher level and try to let \"tertris\" eat more\n"
"than one line at a time.  Polyiamonds with diagonal or "
"with a higher number of triangles earn more points.  Mixed\n"
"mode earns less points.  Play fast and see how well you "
"can do in the bonus mode.  Bonuses are set when the\n"
"level is changed automatically.  (Bonus mode not available "
"for 8 triangles).\n"
"\n"
"The program lists the top 15 players at the end, each player "
"has at most three scores."
};

static const char featuresHelp[] = {
"Click \"mouse-left\" button , or press \"J\", \"j\", "
"\"S\", or \"s\" keys to move the falling polyiamond left.\n"
"\n"
"Click \"mouse-center\" button, or press \"K\", \"k\", "
"\"D\", or \"d\" keys to rotate the falling polyiamond.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"on\".\n"
"\n"
"Click \"mouse-right\" button, or press \"L\", \"l\", "
"\"F\", or \"f\" keys to move the falling polyiamond right.\n"
"\n"
"Click \"mouse-scroll-up\", or press \"I\", \"i\", "
"\"E\", or \"e\" keys to flip (reflect) the falling polyiamond.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"off\".\n"
"\n"
"Click \"mouse-scroll-down\", or press \"space-bar\" "
"to drop the falling polyiamond.\n"
"\n"
"The key pad and arrow keys are defined similar.\n"
"Key pad is defined for Tertris as:\n"
"  /    Reflect\n"
"  8    Rotate\n"
"  ^\n"
"4<5>6  Left, Rotate, Right\n"
"  v\n"
"  2    Drop\n"
"\n"
"Press \"C\" or \"c\" keys to clear.\n"
"\n"
"Press \">\" or \".\" keys to speed up "
"the falling polyiamond.\n"
"\n"
"Press \"<\" or \",\" keys to slow down "
"the falling polyiamond.\n"
"This may be turned off to avoid cheating.\n"
"\n"
"Press \"+\" or \"=\" keys to increase the "
"triangles in the falling polyiamond.  Note: to keep the\n"
"level of play similar, the \"turn style\" can change.  "
"The ability to reflect polyiamonds is available\n"
"only when the number of triangles >= 5.  The ability "
"to rotate polyiamonds, is available only when\n"
"the number of triangles is >= 4.  Bonus polyiamonds "
"still have the current \"turn style\".\n"
"\n"
"Press \"-\" or \"_\" keys to decrease the triangles in "
"the falling polyiamond.\n"
"\n"
"Press \"]\" or \"}\" keys to increase the column/row "
"size.\n"
"\n"
"Press \"[\" or \"{\" keys to decrease the column/row "
"size.\n"
"\n"
"Press \"`\", \"1\", \"2\", \"3\", \"4\", \"5\", "
"\"6\", \"7\", \"8\", \"9\", or \"0\" keys\n"
"(not the keypad 0-9) to change the number of fill lines.  "
"\"`\" for 0 lines and \"0\" for 10 lines.\n"
"\n"
"Press \"X\" or \"x\" to toggle between fixed fill or "
"random fill.\n"
"\n"
"Press \"R\" or \"r\" to toggle between sudden initial "
"appearance or gradual initial appearance.\n"
"\n"
"Press \"G\" or \"g\" to toggle diagonal mode for 2 "
"triangles.\n"
"\n"
"Press \"M\" or \"m\" to toggle mixed mode.  This is "
"ignored for 2 triangles.\n"
"\n"
"Press \"B\" or \"b\" to toggle bonus mode.  This is "
"ignored for 8 triangles.\n"
"\n"
"Press \"W\" or \"w\" to toggle clockwise or counter "
"clockwise turning.\n"
"\n"
"Press \"P\" or \"p\" to toggle plain pieces or "
"three dimensional pieces.\n"
"\n"
"Press \"N\" or \"n\" to toggle show next.\n"
"\n"
"Press \"#\" key to toggle the grid.\n"
"\n"
"Press \"|\" key to pause.\n"
"\n"
"Press \"@\" key to toggle the sound.\n"
"\n"
"Press \"Esc\" key to hide and pause program.  When "
"the tertris window loses its focus, it also goes\n"
"to \"paused\" mode and iconifies itself.\n"
"\n"
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill "
"program."
};

static const char referencesHelp[] = {
"Based at one time on Q. Alex Zhao's Tetris."
};
#endif

#include "file.h"
#ifdef WINVER
#include "TertrisP.h"
#include "TertrisNP.h"
#define TITLE "wtertris"

static TertrisRec widget;
static NextTertrisRec widget2;

#ifndef SCOREPATH
#ifdef UNIXDELIM
#define SCOREPATH "c:/WINDOWS"
#else
#define SCOREPATH "c:\\WINDOWS"
#endif
#endif
#define PRINT_MESSAGE(b) (void) MessageBox(widget.core.hWnd, (LPCSTR) b, "Warning", MB_OK);
#else
#include "xwin.h"
#include "xgui.h"
#include "Tertris.h"
#include "TertrisN.h"
#ifdef HAVE_XPM
#include <X11/xpm.h>
#ifdef CONSTPIXMAPS
#include "16x16/tertris.xpm"
#include "22x22/tertris.xpm"
#include "24x24/tertris.xpm"
#include "32x32/tertris.xpm"
#include "48x48/tertris.xpm"
#include "64x64/tertris.xpm"
#else
#include "pixmaps/16x16/tertris.xpm"
#include "pixmaps/22x22/tertris.xpm"
#include "pixmaps/24x24/tertris.xpm"
#include "pixmaps/32x32/tertris.xpm"
#include "pixmaps/48x48/tertris.xpm"
#include "pixmaps/64x64/tertris.xpm"
#endif
#define RESIZE_XPM(s) ((char **) (((s)<=32)?\
(((s)<=22)?(((s)<=16)?tertris_16x16:tertris_22x22):\
(((s)<=24)?tertris_24x24:tertris_32x32)):\
(((s)<=48)?tertris_48x48:tertris_64x64)))
#endif
#include "pixmaps/64x64/tertris.xbm"
#define DEFINE_XBM (char *) tertris_64x64_bits, tertris_64x64_width, tertris_64x64_height
#ifndef SCOREPATH
#ifdef VMS
#define SCOREPATH "SYS$LOGIN:"
#else
#define SCOREPATH "/var/games/altris"
#endif
#endif
#endif

#ifndef SCOREFILE
#define SCOREFILE "tertris.scores"
#endif

#define NEVER (-1)
#define USER_NAME_LENGTH 120
#define MESSAGE_LENGTH (USER_NAME_LENGTH+64)
#define DATELEN 27
#define MAX_SCORES 15
#define MAX_USER_SCORES 3
#define NOACCESS ""
#define NOBODY "nobody"

typedef struct {
	int score, level, rows;
	char name[USER_NAME_LENGTH]; /* host[USER_NAME_LENGTH] */
	char date[DATELEN];
} GameRecord;

static GameRecord gameRecord[MAX_SCORES];
static int maxScore = 0;
static char messageDsp[MESSAGE_LENGTH] = "Welcome";

#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
static Widget rowsText, scoreText, maxScoreText, message;
static Widget fixedFillToggle, suddenAppearToggle;
static Widget diagonalToggle, mixedToggle, bonusToggle;
static Widget clockwiseToggle, plainPieceToggle, showNextToggle;
static Widget gridToggle;
static Widget soundToggle;
static Widget descriptionDialog, featuresDialog;
static Widget synopsisDialog, referencesDialog, aboutDialog;
static char buff[MESSAGE_LENGTH];
static char *levelLabel = (char *) "Level:";
static char *unitsLabel = (char *) "Units:     ";
static char *fieldLabel = (char *) "Field Size:";
static char *fillLabel = (char *) "Fill:      ";
#define PRINT_MESSAGE(b) printState(message, b)
#elif !defined(WINVER)
#define PRINT_MESSAGE(b) XtWarning(b)
#endif
#ifdef HAVE_MOTIF
static Widget levelChanger;
static Widget unitsChanger, fieldChanger, fillChanger;
static Widget setupDialog = NULL;
static Widget restartDialog;
static Widget menuBar;
#elif defined(HAVE_ATHENA)
static Widget levelSlider, levelSliderLabel;
static Widget unitsSlider, unitsSliderLabel;
static Widget fieldSlider, fieldSliderLabel;
static Widget fillSlider, fillSliderLabel;
static Widget setupDialogShell;
static const char *fileTypes[] =
{
	"Exit"
};
#define numFileTypes (sizeof(fileTypes)/sizeof(fileTypes[0]))
static const char *playTypes[] =
{
	"Pause",
	"Clear",
	"Setup..."
};
#define numPlayTypes (sizeof(playTypes)/sizeof(playTypes[0]))
static const char *helpTypes[] =
{
	"Description...",
	"Features...",
	"Synopsis...",
	"References...",
	"About..."
};
#define numHelpTypes (sizeof(helpTypes)/sizeof(helpTypes[0]))
#else
#ifndef WINVER
static Widget shell;
#endif
#endif
static char scoreFileName[FILE_NAME_LENGTH] = SCOREFILE;
#ifdef WINVER
#define PROGRAM_NAME_LENGTH 80
static char progDsp[PROGRAM_NAME_LENGTH] = TITLE;
static char userNameDsp[USER_NAME_LENGTH] = "guest";
#else
static XtAppContext appCon;
static Pixmap pixmap = None;
static Widget topLevel, tertris, nextTertris;
static char userNameDsp[USER_NAME_LENGTH] = "";

#ifdef HAVE_MOTIF
static void
printState(Widget w, char *msg)
{
	XmString xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("printState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET);
	XtVaSetValues(w,
		XmNlabelString, xmstr, NULL);
	XmStringFree(xmstr);
}
#elif defined(HAVE_ATHENA)
static void
printState(Widget w, char *msg)
{
	if (w == message)
		return;
	XtVaSetValues(w,
		XtNlabel, msg, NULL);
	XtVaSetValues(w,
		XtNwidth, 64, NULL);
}
#endif

static void
printRecords(void)
{
	int i;

	(void) printf("     TERTRIS  HALL OF FAME\n\n");
	(void) printf("   # USER            SCORE LEVEL  ROWS\n");
	for (i = 0; i < MAX_SCORES; i++) {
		if (gameRecord[i].score > 0) {
			(void) printf("%4d %-12s%9d %5d %5d\n", i + 1,
				gameRecord[i].name, gameRecord[i].score,
				gameRecord[i].level, gameRecord[i].rows);
		} else {
			break;
		}
	}
}
#endif

static void
initRecords(void)
{
	int i;

	for (i = 0; i < MAX_SCORES; i++) {
		gameRecord[i].score = NEVER;
		gameRecord[i].level = 0;
		gameRecord[i].rows = 0;
		(void) strncpy(gameRecord[i].name,
			NOACCESS, USER_NAME_LENGTH);
		gameRecord[i].date[0] = '\0';
	}
}

/* #define DEBUG 1 */

static void
readRecords(void)
{
	FILE *fp;
	int s = NEVER, l, r, i;
	char userName[USER_NAME_LENGTH];
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "r")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "r")) == NULL) {
#if 0
			stringCat(&buf1, "Cannot read ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
#endif
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Cannot read ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	free(lname);
	free(fname);
	for (i = 0; i < MAX_SCORES; i++)
		gameRecord[i].score = 0;
	for (i = 0; i < MAX_SCORES; i++) {
		if (fscanf(fp, "%d %d %d %s\n", &s, &l, &r, userName) > 0) {
			gameRecord[i].score = s;
			gameRecord[i].level = l;
			gameRecord[i].rows = r;
			(void) strncpy(gameRecord[i].name,
				userName, USER_NAME_LENGTH);
			if (maxScore < gameRecord[i].score)
				maxScore = gameRecord[i].score;
		} else {
			break;
		}
	}
	(void) fclose(fp);
}

static void
writeRecords(void)
{
	FILE *fp;
	int i;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "w")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Cannot write to ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Cannot write to ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	{
#if HAVE_FCNTL_H
		int lfd;
		char lockFile[FILE_NAME_LENGTH];

		(void) strncpy(lockFile, name, FILE_NAME_LENGTH - 6);
		lockFile[FILE_NAME_LENGTH - 6] = '\0';
		(void) strncat(lockFile, ".lock", 6);
		while (((lfd = open(lockFile, O_CREAT | O_EXCL, 0644)) < 0) &&
				errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr,
				"Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr,
				"Lock file exists... score not recorded - sorry.\n");
			return;
#endif
		}
#endif
		for (i = 0; i < MAX_SCORES; i++) {
			if (gameRecord[i].score <= 0)
				break;
			(void) fprintf(fp, "%d %d %d %s\n",
				gameRecord[i].score,
				gameRecord[i].level,
				gameRecord[i].rows,
				gameRecord[i].name);
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockFile);
#endif
		(void) fclose(fp);
	}
	free(lname);
	free(fname);
}

#ifndef WINVER
/* There is probably a better way to assure that they are the same
 * but I do not know it off hand. */
static void
makeEquivalent(String *userName, String *scoreFile, int *level,
		int *units, Boolean *diagonal, Boolean *mixed, Boolean *bonus,
		Boolean *showNext)
{
	Boolean mono, reverse, plainPiece, grid;
	Boolean scoreOnly, versionOnly;
	Pixel foreground, background;

	XtVaGetValues(tertris,
		XtNuserName, userName,
		XtNscoreFile, scoreFile,
		XtNlevel, level,
		XtNunits, units,
		XtNdiagonal, diagonal,
		XtNmixed, mixed,
		XtNbonus, bonus,
		XtNplainPiece, &plainPiece,
		XtNshowNext, showNext,
		XtNgrid, &grid,
		XtNmono, &mono,
		XtNreverseVideo, &reverse,
		XtNforeground, &foreground,
		XtNbackground, &background,
		XtNscoreOnly, &scoreOnly,
		XtNversionOnly, &versionOnly, NULL);
	if (versionOnly) {
		(void) printf("%s\n", aboutHelp);
		exit(0);
	}
	if (scoreOnly) {
		initRecords();
		readRecords();
		printRecords();
		exit(0);
	}
	XtVaSetValues(nextTertris,
		XtNlevel, *level,
		XtNunits, *units,
		XtNdiagonal, *diagonal,
		XtNmixed, *mixed,
		XtNbonus, *bonus,
		XtNplainPiece, plainPiece,
		XtNshowNext, *showNext,
		XtNgrid, grid,
		XtNmono, mono,
		XtNreverseVideo, reverse,
		XtNforeground, foreground,
		XtNbackground, background, NULL);
}
#endif

static void
printStatus(int numberRows, int yourScore
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
	, char *msg, int level
#endif
	)
{
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
	char titleDsp[TITLE_LENGTH] = "";
#endif
	if (maxScore < yourScore) {
		maxScore = yourScore;
	}
#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
	intCpy(buff, numberRows);
	printState(rowsText, buff);
	intCpy(buff, yourScore);
	printState(scoreText, buff);
	intCpy(buff, maxScore);
	printState(maxScoreText, buff);
#else
#ifdef HAVE_SNPRINTF
	(void) snprintf(titleDsp, TITLE_LENGTH,
		"%s: %d level %d rows @ (%d/%d) - %s",
		progDsp, level, numberRows, yourScore, maxScore, msg);
#else
	(void) sprintf(titleDsp,
		"%s: %d level %d rows @ (%d/%d) - %s",
		progDsp, level, numberRows, yourScore, maxScore, msg);
#endif
#ifdef WINVER
	SetWindowText(widget.core.hWnd, (LPSTR) titleDsp);
#else
	XtVaSetValues(XtParent(tertris),
		XtNtitle, titleDsp, NULL);
#endif
#endif
}

#define CPSCORE(a,b) a.score = b.score, a.level = b.level, a.rows = b.rows,\
	(void) strncpy(a.name, b.name, USER_NAME_LENGTH)

static Boolean
handleGameOver(int lastLevel, int numberRows, int yourScore)
{
	int i, j = 0, k = 0, rank = MAX_SCORES;
	GameRecord current, temp[MAX_SCORES];
	Boolean newRecord = False;

	readRecords();	/* Maybe its been updated by another */
	current.score = yourScore;
	current.level = lastLevel;
	current.rows = numberRows;
	(void) strncpy(current.name, userNameDsp, USER_NAME_LENGTH);
	for (i = 0; i < MAX_SCORES; i++) {
		if (j == 0 && current.score > gameRecord[i].score) {
			CPSCORE(temp[i], current);
			j = 1;
			rank = i;
		} else {
			CPSCORE(temp[i], gameRecord[i - j]);
		}
	}
	j = 0;
	if (rank < MAX_SCORES) {
		for (i = 0; i < MAX_SCORES; i++) {
			if (!strcmp(temp[i].name, userNameDsp)) {
				if (j < MAX_USER_SCORES) {
					CPSCORE(gameRecord[k], temp[i]);
					if (i == rank)
						newRecord = True;
					k++;
				}
				j++;
			} else {
				CPSCORE(gameRecord[k], temp[i]);
				k++;
			}
		}
		for (i = k; i < MAX_SCORES; i++) {
			gameRecord[i].score = 0;
		}
	}
	if (newRecord) {
		writeRecords();
		printStatus(numberRows, yourScore
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
			, messageDsp, lastLevel
#endif
			);
		return True;
	}
	return False;
}

#ifdef WINVER
static LRESULT CALLBACK
showNextCallback(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	PAINTSTRUCT paint;

	if (widget2.core.hWnd == NULL)
		widget2.core.hWnd = hDlg;
	if (GetFocus()) {
		if (!widget2.tertris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_BRUSH));
			enterNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	} else {
		if (widget2.tertris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_BRUSH));
			leaveNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	}
	switch (message) {
	case WM_DESTROY:
		destroyNextGame(brush);
		(void) DestroyWindow(widget2.core.hWnd);
		widget2.core.hWnd = NULL;
		return TRUE;
	case WM_SIZE:
		resizeNextGame(&widget2);
		(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
		break;
	case WM_PAINT:
		exposeNextGame(&widget2);
		(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case ACTION_EXIT:
			destroyNextGame(brush);
			(void) DestroyWindow(widget2.core.hWnd);
			widget2.core.hWnd = NULL;
			return TRUE;
		case ACTION_HIDE:
			hideNextGame(&widget2);
			break;
#if 0
		case ACTION_CLEAR:
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_PEN));
			showNextGame(&widget2);
			(void) ReleaseDC(widget2.core.hWnd, widget2.core.hDC);
			break;
#endif
		case ACTION_SHOW_NEXT:
			if (widget2.tertris.showNext) {
				ShowWindow(widget2.core.hWnd,
					SW_SHOWNOACTIVATE);
			} else {
				ShowWindow(widget2.core.hWnd, SW_HIDE);
			}
			break;
#if 0
		case ACTION_SHOW_NEXT:
			if (widget2.core.hWnd == NULL)
				widget2.core.hWnd =
					CreateDialog(widget2.core.hInstance,
					"ShowNext", widget2.core.hWnd,
					(DLGPROC) showNextCallback);
			else {
				destroyNextGame(brush);
				(void) DestroyWindow(widget2.core.hWnd);
				widget2.core.hWnd = NULL;
				return TRUE;
			}
			break;
#endif
		}
		break;
	/* default:
		return (DefWindowProc(widget2.core.hWnd, message,
			wParam, lParam)); */
	}
	return FALSE;
}
#endif

static void
initialize(
#ifdef WINVER
TertrisWidget w, HBRUSH brush, HBRUSH brush2
#else
Widget w
#endif
)
{
	int level;
	char *userName, *scoreFile;

#ifdef WINVER
	initializeGame(w, brush);
	level = w->tertris.level;
	/*units = */widget2.tertris.units = w->tertris.units;
	/*diagonal = */widget2.tertris.diagonal = w->tertris.diagonal;
	/*mixed = */widget2.tertris.mixed = w->tertris.mixed;
	/*bonus = */widget2.tertris.bonus = w->tertris.bonus;
	userName = w->tertris.userName;
	scoreFile = w->tertris.scoreFile;
	initializeNextGame(&widget2, brush2);
	(void) CreateDialog(widget.core.hInstance,
		"ShowNext", widget.core.hWnd, (DLGPROC) showNextCallback);
	if (!widget2.tertris.showNext)
		ShowWindow(widget2.core.hWnd, SW_HIDE);
	/*showNext = */widget.tertris.showNext = widget2.tertris.showNext;
#else
	int units;
	Boolean diagonal, mixed, bonus, showNext;

	makeEquivalent(&userName, &scoreFile, &level, &units,
		&diagonal, &mixed, &bonus, &showNext);
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
	XtVaSetValues(levelChanger,
		XmNposition, level, NULL);
#else
	XmScaleSetValue(levelChanger, level);
#endif
#elif defined(HAVE_ATHENA)
	setScale(levelSlider, levelSliderLabel, level,
		MIN_LEVELS, MAX_LEVELS, True);
	XtVaSetValues(diagonalToggle,
		XtNstate, diagonal, NULL);
	XtVaSetValues(mixedToggle,
		XtNstate, mixed, NULL);
	XtVaSetValues(bonusToggle,
		XtNstate, bonus, NULL);
	XtVaSetValues(showNextToggle,
		XtNstate, showNext, NULL);
	{
		Boolean fixedFill, suddenAppear;
		Boolean cw, plainPiece, grid, sound;
		XtVaGetValues(tertris,
			XtNfixedFill, &fixedFill,
			XtNsuddenAppear, &suddenAppear,
			XtNcw, &cw,
			XtNplainPiece, &plainPiece,
			XtNgrid, &grid,
			XtNsound, &sound, NULL);
		XtVaSetValues(fixedFillToggle,
			XtNstate, fixedFill, NULL);
		XtVaSetValues(suddenAppearToggle,
			XtNstate, suddenAppear, NULL);
		XtVaSetValues(clockwiseToggle,
			XtNstate, cw, NULL);
		XtVaSetValues(plainPieceToggle,
			XtNstate, plainPiece, NULL);
		XtVaSetValues(gridToggle,
			XtNstate, grid, NULL);
		XtVaSetValues(soundToggle,
			XtNstate, sound, NULL);
	}
#endif
#endif
	initRecords();
	readRecords();
#ifndef WINVER
	(void) strncpy(userNameDsp, userName, USER_NAME_LENGTH - 1);
	userNameDsp[USER_NAME_LENGTH - 1] = '\0';
#endif
	if (!strcmp(userName, "(null)") || !strcmp(userName, NOBODY) ||
			!strcmp(userName, NOACCESS)) {
#ifdef WINVER
		(void) strncpy(userNameDsp, userName, USER_NAME_LENGTH - 1);
		userNameDsp[USER_NAME_LENGTH - 1] = '\0';
#else
		char *login;

#ifdef HAVE_CUSERID
		login = cuserid(NULL);
#else
		login = getlogin();
#endif
		if (login == NULL) {
			userNameDsp[0] = '\0';
		} else {
			(void) strncpy(userNameDsp, login, USER_NAME_LENGTH - 1);
			userNameDsp[USER_NAME_LENGTH - 1] = '\0';
		}
		if (!strcmp(userNameDsp, "(null)") ||
				!strcmp(userNameDsp, NOBODY) ||
				!strcmp(userNameDsp, NOACCESS))
			/* It really IS nobody */
			(void) strncpy(userNameDsp, "guest", 6);
#endif
	}
	if (strcmp(scoreFile, "")) {
		(void) strncpy(scoreFileName, scoreFile, FILE_NAME_LENGTH - 1);
		scoreFileName[FILE_NAME_LENGTH - 1] = '\0';
	}
	printStatus(0, 0
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
		, messageDsp, level
#endif
		);
}

#ifdef WINVER
void
setGame(TertrisWidget w, int reason)
#else
static void
callbackGame(Widget w, caddr_t clientData, tertrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif
	int level, units, columns, fill, score, completedRows;
	Boolean fixedFill, suddenAppear, diagonal, mixed, bonus;
	Boolean cw, plainPiece, grid, showNext, sound;

	messageDsp[0] = '\0';
#ifdef WINVER
	level = w->tertris.level;
	units = w->tertris.units;
	columns = w->tertris.columns;
	fill = w->tertris.fill;
	fixedFill = w->tertris.fixedFill;
	suddenAppear = w->tertris.suddenAppear;
	diagonal = w->tertris.nextSpec.diagonal;
	mixed = w->tertris.nextSpec.mixed;
	bonus = w->tertris.bonus;
	cw = w->tertris.cw;
	plainPiece = w->tertris.plainPiece;
	grid = w->tertris.grid;
	showNext = w->tertris.showNext;
	sound = w->tertris.sound;
	completedRows = w->tertris.completedRows;
	score = w->tertris.score;
#else
	XtVaGetValues(w,
		XtNlevel, &level,
		XtNunits, &units,
		XtNcolumns, &columns,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNgrid, &grid,
		XtNsound, &sound,
		XtNcompletedRows, &completedRows,
		XtNscore, &score, NULL);
	XtVaGetValues(nextTertris,
		XtNshowNext, &showNext, NULL);
#endif
	switch (reason) {
	case ACTION_HIDE:
#ifdef WINVER
		ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
		(void) XIconifyWindow(XtDisplay(topLevel),
			XtWindow(topLevel),
			XScreenNumberOfScreen(XtScreen(topLevel)));
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
		(void) XIconifyWindow(XtDisplay(shell),
			XtWindow(shell),
			XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
		break;
#ifndef WINVER
	case ACTION_RESTART_QUERY:
#ifdef HAVE_MOTIF
		XtManageChild(restartDialog);
#else
		XtVaSetValues(tertris,
			XtNmenu, 4, NULL); /* menu choice */
#endif
		break;
#endif
	case ACTION_STATUS:
		printStatus(completedRows, score
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
			, messageDsp, level
#endif
			);
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
		XtVaSetValues(levelChanger,
			XmNposition, level, NULL);
#else
		XmScaleSetValue(levelChanger, level);
#endif
#endif
		break;
#ifndef WINVER
	case ACTION_DRAW_NEXT:
		XtVaSetValues(nextTertris,
			XtNrandomNumber, callData->randomNumber,
			XtNrandomRotation, callData->randomRotation,
			XtNrandomReflection, callData->randomReflection,
			XtNunits, callData->units,
			XtNdiagonal, callData->diagonal,
			XtNmixed, callData->mixed,
			XtNbonus, callData->newBonus, NULL);
		break;
#endif
	case ACTION_SHOW_NEXT:
		showNext = !showNext;
#ifdef WINVER
		showNextNextGame(&widget2);
#else
		XtVaSetValues(nextTertris,
			XtNshowNext, showNext, NULL);
#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
		XtVaSetValues(nextTertris,
			XtNrandomNumber, CLEARNEXT_RAND, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(showNextToggle,
				showNext, True);
#endif
#else
		if (showNext) {
			int x, y, width;

			XtVaGetValues(topLevel,
				XtNx, &x,
				XtNy, &y,
				XtNwidth, &width, NULL);
			XtVaSetValues(shell,
				XtNx, x + width,
				XtNy, y, NULL);
			XtRealizeWidget(shell);
			XtVaSetValues(nextTertris,
				XtNrandomNumber, CLEARNEXT_RAND, NULL);
		} else {
			XtVaSetValues(nextTertris,
				XtNrandomNumber, CLEARNEXT_RAND, NULL);
			XtUnrealizeWidget(shell);
		}
#endif
#endif
		break;
	case ACTION_DEC_UNITS:
		units--;
#ifdef WINVER
		w->tertris.units = units;
#else
		XtVaSetValues(tertris,
			XtNunits, units, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL) {
#ifdef USE_SPIN
			XtVaSetValues(unitsChanger,
				XmNposition, units, NULL);
#else
			XmScaleSetValue(unitsChanger, units);
#endif
		}
#endif
#endif
		break;
	case ACTION_INC_UNITS:
		units++;
#ifdef WINVER
		w->tertris.units = units;
#else
		XtVaSetValues(tertris,
			XtNunits, units, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL) {
#ifdef USE_SPIN
			XtVaSetValues(unitsChanger,
				XmNposition, units, NULL);
#else
			XmScaleSetValue(unitsChanger, units);
#endif
		}
#endif
#endif
		break;
	case ACTION_DEC_COLUMNS:
		columns -= 2;
#ifdef WINVER
		w->tertris.columns = columns;
#else
		XtVaSetValues(tertris,
			XtNcolumns, columns, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL) {
#ifdef USE_SPIN
			XtVaSetValues(fieldChanger,
				XmNposition, columns, NULL);
#else
			XmScaleSetValue(fieldChanger, columns);
#endif
		}
#endif
#endif
		break;
	case ACTION_INC_COLUMNS:
		columns += 2;
#ifdef WINVER
		w->tertris.columns = columns;
#else
		XtVaSetValues(tertris,
			XtNcolumns, columns, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL) {
#ifdef USE_SPIN
			XtVaSetValues(fieldChanger,
				XmNposition, columns, NULL);
#else
			XmScaleSetValue(fieldChanger, columns);
#endif
		}
#endif
#endif
		break;
	case ACTION_FIXED_FILL:
		fixedFill = !fixedFill;
#ifdef WINVER
		w->tertris.fixedFill = fixedFill;
#else
		XtVaSetValues(tertris,
			XtNfixedFill, fixedFill, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(fixedFillToggle,
				fixedFill, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(fixedFillToggle,
			XtNstate, fixedFill, NULL);
#endif
#endif
		break;
	case ACTION_SUDDEN_APPEAR:
		suddenAppear = !suddenAppear;
#ifdef WINVER
		w->tertris.suddenAppear = suddenAppear;
#else
		XtVaSetValues(tertris,
			XtNsuddenAppear, suddenAppear, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(suddenAppearToggle,
				suddenAppear, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(suddenAppearToggle,
			XtNstate, suddenAppear, NULL);
#endif
#endif
		break;
	case ACTION_DIAGONAL:
		diagonal = !diagonal;
#ifdef WINVER
		w->tertris.diagonal = diagonal;
#else
		XtVaSetValues(tertris,
			XtNdiagonal, diagonal, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(diagonalToggle,
				diagonal, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(diagonalToggle,
			XtNstate, diagonal, NULL);
#endif
#endif
		break;
	case ACTION_MIXED:
		mixed = !mixed;
#ifdef WINVER
		w->tertris.mixed = mixed;
#else
		XtVaSetValues(tertris,
			XtNmixed, mixed, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(mixedToggle,
				mixed, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(mixedToggle,
			XtNstate, mixed, NULL);
#endif
#endif
		break;
	case ACTION_BONUS:
		bonus = !bonus;
#ifdef WINVER
		w->tertris.bonus = bonus;
#else
		XtVaSetValues(tertris,
			XtNbonus, bonus, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(bonusToggle,
				bonus, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(bonusToggle,
			XtNstate, bonus, NULL);
#endif
#endif
		break;
	case ACTION_CW:
		cw = !cw;
#ifdef WINVER
		w->tertris.cw = cw;
#else
		XtVaSetValues(tertris,
			XtNcw, cw, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(clockwiseToggle, cw, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(clockwiseToggle,
			XtNstate, cw, NULL);
#endif
#endif
		break;
	case ACTION_PLAIN_PIECE:
		plainPiece = !plainPiece;
#ifdef WINVER
		w->tertris.plainPiece = plainPiece;
		widget2.tertris.plainPiece = plainPiece;
#else
		XtVaSetValues(tertris,
			XtNplainPiece, plainPiece, NULL);
		XtVaSetValues(nextTertris,
			XtNplainPiece, plainPiece, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(plainPieceToggle, plainPiece,
				True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(plainPieceToggle,
			XtNstate, plainPiece, NULL);
#endif
#endif
		break;
	case ACTION_GRID:
		grid = !grid;
#ifdef WINVER
		w->tertris.grid = grid;
		widget2.tertris.grid = grid;
#else
		XtVaSetValues(tertris,
			XtNgrid, grid, NULL);
		XtVaSetValues(nextTertris,
			XtNgrid, grid, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(gridToggle, grid, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(gridToggle,
			XtNstate, grid, NULL);
#endif
#endif
		break;
	case ACTION_SOUND:
		sound = !sound;
#ifdef WINVER
		w->tertris.sound = sound;
#else
		XtVaSetValues(tertris,
			XtNsound, sound, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(soundToggle,
				sound, True);
#elif defined(HAVE_ATHENA)
		XtVaSetValues(soundToggle,
			XtNstate, sound, NULL);
#endif
#endif
		break;
	case ACTION_FILL0:
	case ACTION_FILL1:
	case ACTION_FILL2:
	case ACTION_FILL3:
	case ACTION_FILL4:
	case ACTION_FILL5:
	case ACTION_FILL6:
	case ACTION_FILL7:
	case ACTION_FILL8:
	case ACTION_FILL9:
	case ACTION_FILL10:
		fill = reason - ACTION_FILL0;
#ifdef WINVER
		w->tertris.fill = fill;
#else
		XtVaSetValues(tertris,
			XtNfill, fill, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL) {
#ifdef USE_SPIN
			XtVaSetValues(fillChanger,
				XmNposition, fill, NULL);
#else
			XmScaleSetValue(fillChanger, fill);
#endif
		}
#endif
#endif
		break;
	case ACTION_GAME_OVER:
		if (handleGameOver(level, completedRows, score)) {
			(void) strncpy(messageDsp, "Congratulations ", 17);
			(void) strncat(messageDsp, userNameDsp, USER_NAME_LENGTH);
			(void) strncat(messageDsp, "!!", 3);
		} else {
			(void) strncpy(messageDsp, "Game Over!", 11);
		}
		break;
	}
#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
	printState(message, messageDsp);
#endif
}

#ifdef WINVER
void
setGameNext(TertrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	if (reason == ACTION_DRAW_NEXT) {
		valueNextGame(&widget2,
			randomNumber, randomRotation, randomReflection,
			units, diagonal, mixed, newBonus);
	}
}
#endif

#ifdef WINVER
void
setNextGame(NextTertrisWidget w, int reason)
#else
static void
callbackNextGame(Widget w, caddr_t clientData, tertrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif

	switch (reason) {
	case ACTION_HIDE:
#ifdef WINVER
		ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
		(void) XIconifyWindow(XtDisplay(topLevel),
			XtWindow(topLevel),
			XScreenNumberOfScreen(XtScreen(topLevel)));
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
		(void) XIconifyWindow(XtDisplay(shell),
			XtWindow(shell),
			XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
		break;
	}
}

#ifdef WINVER
static LRESULT CALLBACK
about(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_COMMAND && LOWORD(wParam) == IDOK) {
		(void) EndDialog(hDlg, TRUE);
		return TRUE;
	}
	return FALSE;
}

static LRESULT CALLBACK
WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	HBRUSH brush2 = (HBRUSH) NULL;
	PAINTSTRUCT paint;
	static int count = 0;

	widget.core.hWnd = hWnd;
	if (GetFocus()) {
		if (!widget.tertris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_BRUSH));
			enterGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	} else {
		if (widget.tertris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_BRUSH));
			leaveGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	}
	switch (message) {
	case WM_CREATE:
		initialize(&widget, brush, brush2);
		break;
	case WM_DESTROY:
		destroyGame(brush);
		break;
	case WM_SIZE:
		resizeGame(&widget);
		(void) InvalidateRect(hWnd, NULL, TRUE);
		break;
	case WM_PAINT:
		widget.core.hDC = BeginPaint(hWnd, &paint);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		exposeGame(&widget);
		(void) EndPaint(hWnd, &paint);
		break;
	case WM_TIMER:
		KillTimer(hWnd, 1);
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		if (count++ == 0) /* kludge to draw first time */
			drawPieceNextGame(&widget2);
		falldownGame(&widget);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		SetTimer(hWnd, 1,
			(unsigned int) (7 * widget.tertris.delay), NULL);
		break;
	case WM_RBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, MOVE_RIGHT);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
	case WM_MBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, ROTATE);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
	case WM_LBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, MOVE_LEFT);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
#if (_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400)
	case WM_MOUSEWHEEL:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		{
			int zDelta = ((short) HIWORD(wParam));

			if (zDelta > (WHEEL_DELTA >> 1)) {
				(void) moveGameInput(&widget, REFLECT);
			} else if (zDelta < -(WHEEL_DELTA >> 1)) {
				(void) moveGameInput(&widget, DROP);
			}
		}
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
#endif
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case ACTION_EXIT:
			destroyGame(brush);
			break;
		case ACTION_HIDE:
			hideGame(&widget);
			break;
		case ACTION_CLEAR:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			clearGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_RIGHT:
		case ACTION_LEFT:
		case ACTION_ROTATE:
		case ACTION_REFLECT:
		case ACTION_DROP:
			{
				widget.core.hDC = GetDC(hWnd);
				(void) SelectObject(widget.core.hDC,
					GetStockObject(NULL_PEN));
				(void) moveGameInput(&widget,
					(int) LOWORD(wParam) - ACTION_DROP + 1);
				(void) ReleaseDC(hWnd, widget.core.hDC);
			}
			break;
		case ACTION_INC_LEVEL:
			(void) incrementGameLevel(&widget);
			break;
		case ACTION_DEC_LEVEL:
			(void) decrementGameLevel(&widget);
			break;
		case ACTION_INC_UNITS:
			(void) incrementGameUnits(&widget);
			break;
		case ACTION_DEC_UNITS:
			(void) decrementGameUnits(&widget);
			break;
		case ACTION_INC_COLUMNS:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) incrementGameColumns(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_DEC_COLUMNS:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) decrementGameColumns(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_FIXED_FILL:
			(void) fixedfillGame(&widget);
			break;
		case ACTION_SUDDEN_APPEAR:
			(void) suddenAppearGame(&widget);
			break;
		case ACTION_DIAGONAL:
			(void) diagonalGame(&widget);
			break;
		case ACTION_MIXED:
			(void) mixedGame(&widget);
			break;
		case ACTION_BONUS:
			(void) bonusGame(&widget);
			break;
		case ACTION_CW:
			(void) cwGame(&widget);
			break;
		case ACTION_PLAIN_PIECE:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) plainPieceGame(&widget);
			drawPieceNextGame(&widget2);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_GRID:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) gridGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_SOUND:
			(void) soundGame(&widget);
			break;
		case ACTION_FILL0:
		case ACTION_FILL1:
		case ACTION_FILL2:
		case ACTION_FILL3:
		case ACTION_FILL4:
		case ACTION_FILL5:
		case ACTION_FILL6:
		case ACTION_FILL7:
		case ACTION_FILL8:
		case ACTION_FILL9:
		case ACTION_FILL10:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) fillGame(&widget,
				(int) LOWORD(wParam) - ACTION_FILL0);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_SHOW_NEXT:
			showNextNextGame(&widget2);
			if (widget2.tertris.showNext) {
				ShowWindow(widget2.core.hWnd,
					SW_SHOWNOACTIVATE);
			} else {
				ShowWindow(widget2.core.hWnd, SW_HIDE);
			}
#if 0
			if (widget2.core.hWnd == NULL) {
				widget2.core.hWnd =
					CreateDialog(widget.core.hInstance,
					"ShowNext", hWnd, (DLGPROC) showNextCallback);
			} else {
				destroyNextGame(brush2);
				(void) DestroyWindow(widget2.core.hWnd);
				widget2.core.hWnd = NULL;
			}
#endif
			break;
		case ACTION_DESCRIPTION:
			(void) MessageBox(hWnd, descriptionHelp,
				"Description", MB_OK | MB_ICONQUESTION);
			break;
		case ACTION_FEATURES:
			(void) MessageBox(hWnd, featuresHelp,
				"Features", MB_OK | MB_ICONEXCLAMATION);
			break;
		case ACTION_REFERENCES:
			(void) MessageBox(hWnd, referencesHelp,
				"References", MB_OK | MB_ICONINFORMATION);
			break;
		case ACTION_ABOUT:
			(void) DialogBox(widget.core.hInstance,
				"About", hWnd, (DLGPROC) about);
			break;
		}
		break;
	default:
		return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	return FALSE;
}

int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
		int numCmdShow)
{
	HWND hWnd;
	MSG msg;
	WNDCLASS wc;
	HACCEL hAccel;

	if (!hPrevInstance) {
		wc.style = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc = WindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = LoadIcon(hInstance, TITLE);
		wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH) GetStockObject(GRAY_BRUSH);
		wc.lpszMenuName = TITLE;
		wc.lpszClassName = TITLE;
		if (!RegisterClass(&wc))
			return FALSE;
	}
	widget.core.hInstance = hInstance;
	hWnd = CreateWindow(TITLE,
		TITLE,
		WS_OVERLAPPEDWINDOW,
		(signed) CW_USEDEFAULT,
		(signed) CW_USEDEFAULT,
		289,
		655,
		HWND_DESKTOP,
		(HMENU) NULL,
		hInstance,
		(XtPointer) NULL);
	if (!hWnd)
		return FALSE;
	hAccel = (HACCEL) LoadAccelerators(hInstance, TITLE);
	(void) ShowWindow(hWnd, numCmdShow);
	SetTimer(hWnd, 1, 1000, NULL);
	(void) UpdateWindow(hWnd);
	while (GetMessage(&msg, (HWND) NULL, 0, 0))
		if (!IsDialogMessage(widget2.core.hWnd, &msg)) {
			if (!TranslateAccelerator(hWnd, hAccel, &msg)) {
				(void) TranslateMessage(&msg);
				(void) DispatchMessage(&msg);
			}
		}
	KillTimer(hWnd, 1);
	return (int) msg.wParam;
}

#else

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*tertris.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*tertris.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*tertris.reverseVideo", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*tertris.reverseVideo", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*tertris.reverseVideo", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*tertris.reverseVideo", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*tertris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*tertris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-level", (char *) "*tertris.level", XrmoptionSepArg, NULL},
	{(char *) "-units", (char *) "*tertris.units", XrmoptionSepArg, NULL},
	{(char *) "-columns", (char *) "*tertris.columns", XrmoptionSepArg, NULL},
	{(char *) "-fill", (char *) "*tertris.fill", XrmoptionSepArg, NULL},
	{(char *) "-fixedFill", (char *) "*tertris.fixedFill", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nofixedFill", (char *) "*tertris.fixedFill", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-suddenAppear", (char *) "*tertris.suddenAppear", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosuddenAppear", (char *) "*tertris.suddenAppear", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-diagonal", (char *) "*tertris.diagonal", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nodiagonal", (char *) "*tertris.diagonal", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-mixed", (char *) "*tertris.mixed", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomixed", (char *) "*tertris.mixed", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bonus", (char *) "*tertris.bonus", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nobonus", (char *) "*tertris.bonus", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-cw", (char *) "*tertris.cw", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nocw", (char *) "*tertris.cw", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-plain", (char *) "*tertris.plain", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noplain", (char *) "*tertris.plain", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-showNext", (char *) "*tertris.showNext", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noshowNext", (char *) "*tertris.showNext", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-grid", (char *) "*tertris.grid", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nogrid", (char *) "*tertris.grid", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-sound", (char *) "*tertris.sound", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosound", (char *) "*tertris.sound", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bumpSound", (char *) "*tertris.bumpSound", XrmoptionSepArg, NULL},
	{(char *) "-userName", (char *) "*tertris.userName", XrmoptionSepArg, NULL},
	{(char *) "-scoreFile", (char *) "*tertris.scoreFile", XrmoptionSepArg,
 NULL},
	{(char *) "-scores", (char *) "*tertris.scoreOnly", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-version", (char *) "*tertris.versionOnly", XrmoptionNoArg, (char *) "TRUE"}
};

static void
playingTimer(XtPointer clientData, XtIntervalId *id)
{
	int delay;

	XtVaGetValues(tertris,
		XtNdelay, &delay, NULL);
	XtVaSetValues(tertris,
		XtNact, True, NULL);
	(void) XtAppAddTimeOut(appCon, 7 * delay, playingTimer,
		clientData);
}

#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
#if defined(HAVE_ATHENA) && defined(USE_SPIN)
static void
unitsUpCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int units;

	XtVaGetValues(tertris,
		XtNunits, &units, NULL);
	units++;
	if (units > MAX_UNITS) {
		return;
	}
	setScale(unitsSlider, unitsSliderLabel, units,
		MIN_UNITS, MAX_UNITS, True);
	XtVaSetValues(tertris,
		XtNunits, units, NULL);
}

static void
unitsDownCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int units;

	XtVaGetValues(tertris,
		XtNunits, &units, NULL);
	units--;
	if (units < MIN_UNITS) {
		return;
	}
	setScale(unitsSlider, unitsSliderLabel, units,
		MIN_UNITS, MAX_UNITS, True);
	XtVaSetValues(tertris,
		XtNunits, units, NULL);
}

#else

static void
unitsChangeCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
		XmSpinBoxCallbackStruct *cbs
#else
		XmScaleCallbackStruct *cbs
#endif
#elif defined(HAVE_ATHENA)
		XtPointer callData
#endif
		)
{
	int units, old;

#ifdef HAVE_MOTIF
#ifdef USE_SPIN
	units = cbs->position;
#else
	units = cbs->value;
#endif
#elif defined(HAVE_ATHENA)
	Widget label = (Widget) clientData;
	if (!thumbScroll(callData, &units,
			MIN_UNITS, MAX_UNITS, SCROLL_SIZE)) {
		return;
	}
#endif
	XtVaGetValues(tertris,
		XtNunits, &old, NULL);
	if (old == units)
		return;
#ifdef HAVE_ATHENA
	setScale(w, label, units, MIN_UNITS, MAX_UNITS, True);
#endif
	XtVaSetValues(tertris,
		XtNunits, units, NULL);
}
#endif

#if defined(HAVE_ATHENA) && defined(USE_SPIN)
static void
fieldUpCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int columns;

	XtVaGetValues(tertris,
		XtNcolumns, &columns, NULL);
	columns++;
	if (columns > MAX_COLUMNS) {
		return;
	}
	setScale(fieldSlider, fieldSliderLabel, columns,
		MIN_COLUMNS, MAX_COLUMNS, True);
	XtVaSetValues(tertris,
		XtNcolumns, columns, NULL);
}

static void
fieldDownCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int columns;

	XtVaGetValues(tertris,
		XtNcolumns, &columns, NULL);
	columns--;
	if (columns < MIN_COLUMNS) {
		return;
	}
	setScale(fieldSlider, fieldSliderLabel, columns,
		MIN_COLUMNS, MAX_COLUMNS, True);
	XtVaSetValues(tertris,
		XtNcolumns, columns, NULL);
}

#else

static void
fieldChangeCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
		XmSpinBoxCallbackStruct *cbs
#else
		XmScaleCallbackStruct *cbs
#endif
#elif defined(HAVE_ATHENA)
		XtPointer callData
#endif
		)
{
	int columns, old;

#ifdef HAVE_MOTIF
#ifdef USE_SPIN
	columns = cbs->position;
#else
	columns = cbs->value;
#endif
#elif defined(HAVE_ATHENA)
	Widget label = (Widget) clientData;
	if (!thumbScroll(callData, &columns,
			MIN_COLUMNS, MAX_COLUMNS, SCROLL_SIZE)) {
		return;
	}
#endif
	XtVaGetValues(tertris,
		XtNcolumns, &old, NULL);
	if (old == columns)
		return;
#ifdef HAVE_ATHENA
	setScale(w, label, columns, MIN_COLUMNS, MAX_COLUMNS, True);
#endif
	XtVaSetValues(tertris,
		XtNcolumns, columns, NULL);
}
#endif

#if defined(HAVE_ATHENA) && defined(USE_SPIN)
static void
fillUpCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int fill;

	XtVaGetValues(tertris,
		XtNfill, &fill, NULL);
	fill++;
	if (fill > MAX_FILL) {
		return;
	}
	setScale(fillSlider, fillSliderLabel, fill,
		0, MAX_FILL, True);
	XtVaSetValues(tertris,
		XtNfill, fill, NULL);
}

static void
fillDownCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int fill;

	XtVaGetValues(tertris,
		XtNfill, &fill, NULL);
	fill--;
	if (fill < 0) {
		return;
	}
	setScale(fillSlider, fillSliderLabel, fill,
		0, MAX_FILL, True);
	XtVaSetValues(tertris,
		XtNfill, fill, NULL);
}

#else

static void
fillChangeCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
		XmSpinBoxCallbackStruct *cbs
#else
		XmScaleCallbackStruct *cbs
#endif
#elif defined(HAVE_ATHENA)
		XtPointer callData
#endif
		)
{
	int fill, old;

#ifdef HAVE_MOTIF
#ifdef USE_SPIN
	fill = cbs->position;
#else
	fill = cbs->value;
#endif
#elif defined(HAVE_ATHENA)
	Widget label = (Widget) clientData;
	if (!thumbScroll(callData, &fill,
			0, MAX_FILL, SCROLL_SIZE)) {
		return;
	}
#endif
	XtVaGetValues(tertris,
		XtNfill, &old, NULL);
	if (old == fill)
		return;
#ifdef HAVE_ATHENA
	setScale(w, label, fill, 0, MAX_FILL, True);
#endif
	XtVaSetValues(tertris,
		XtNfill, fill, NULL);
}
#endif

#if defined(HAVE_ATHENA) && defined(USE_SPIN)
static void
levelUpCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int level;

	XtVaGetValues(tertris,
		XtNlevel, &level, NULL);
	level++;
	if (level > MAX_LEVELS) {
		return;
	}
	setScale(levelSlider, levelSliderLabel, level,
		MIN_LEVELS, MAX_LEVELS, True);
	XtVaSetValues(tertris,
		XtNlevel, level, NULL);
}

static void
levelDownCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int level;

	XtVaGetValues(tertris,
		XtNlevel, &level, NULL);
	level--;
	if (level < MIN_LEVELS) {
		return;
	}
	setScale(levelSlider, levelSliderLabel, level,
		MIN_LEVELS, MAX_LEVELS, True);
	XtVaSetValues(tertris,
		XtNlevel, level, NULL);
}

#else

static void
levelChangeCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
#ifdef USE_SPIN
		XmSpinBoxCallbackStruct *cbs
#else
		XmScaleCallbackStruct *cbs
#endif
#elif defined(HAVE_ATHENA)
		XtPointer callData
#endif
		)
{
	int level, old;

#ifdef HAVE_MOTIF
#ifdef USE_SPIN
	level = cbs->position;
#else
	level = cbs->value;
#endif
#elif defined(HAVE_ATHENA)
	Widget label = (Widget) clientData;
	if (!thumbScroll(callData, &level,
			MIN_LEVELS, MAX_LEVELS, SCROLL_SIZE)) {
		return;
	}
#endif
	XtVaGetValues(tertris,
		XtNlevel, &old, NULL);
	if (old == level)
		return;
#ifdef HAVE_ATHENA
	setScale(w, label, level, MIN_LEVELS, MAX_LEVELS, True);
#endif
	XtVaSetValues(tertris,
		XtNlevel, level, NULL);
}
#endif

static void
fixedFillCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNfixedFill, state, NULL);
}

static void
suddenAppearCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNsuddenAppear, state, NULL);
}

static void
diagonalCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNdiagonal, state, NULL);
}

static void
mixedCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNmixed, state, NULL);
}

static void
bonusCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNbonus, state, NULL);
}

static void
clockwiseCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNcw, state, NULL);
}

static void
plainPieceCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNplainPiece, state, NULL);
	XtVaSetValues(nextTertris,
		XtNplainPiece, state, NULL);
}

static void
showNextCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(nextTertris,
		XtNshowNext, state, NULL);
}

static void
gridCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNgrid, state, NULL);
	XtVaSetValues(nextTertris,
		XtNgrid, state, NULL);
}

static void
soundCallback(Widget w, XtPointer clientData,
#ifdef HAVE_MOTIF
		XmToggleButtonCallbackStruct *cbs
#elif defined(HAVE_ATHENA)
		int state
#endif
		)
{
#ifdef HAVE_MOTIF
	Boolean state = cbs->set;
#endif
	XtVaSetValues(tertris,
		XtNsound, state, NULL);
}

static void
fileMenuCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int value = (int) ((size_t) clientData);

	if (value == 0)
		exit(0);
}

#ifdef HAVE_MOTIF
static Widget createSetup(Widget w, char *title);
#elif defined(HAVE_ATHENA)
static void
createSetupDialog(Widget w)
/* , XtPointer clientData, XtPointer callData)*/
{
	/*Widget topLevel = (Widget) clientData;*/
	Position x, y;
	Dimension width, height;

	if (wmDeleteWindow == None)
		wmDeleteWindow = XInternAtom(XtDisplay(topLevel),
			"WM_DELETE_WINDOW", FALSE);
	XtVaGetValues(topLevel,
		XtNwidth, &width,
		XtNheight, &height, NULL);
	/*
	 * translate coordinates in application top-level window
	 * into coordinates from root window origin.
	 */
	XtTranslateCoords(topLevel,
		(Position) (width >> 1), /* x */
		(Position) (height >> 1), /* y */
		&x, &y); /* coords on root window */
	/* move popup shell to this position (it's not visible yet) */
	XtVaSetValues(w,
		XtNx, x,
		XtNy, y, NULL);
}
#endif

static void
playMenuCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int value = (int) ((size_t) clientData);

	if (value == 2) {
#ifdef HAVE_MOTIF
		/*XtManageChild(setupDialog); */
		if (setupDialog == NULL)
			setupDialog = createSetup(menuBar, (char *) "Setup");
#elif defined(HAVE_ATHENA)
		createSetupDialog(setupDialogShell);
		XtPopup(setupDialogShell, XtGrabNone);
		XSetWMProtocols(XtDisplay(topLevel),
			XtWindow(setupDialogShell), &wmDeleteWindow, 1);
#endif
		return;
	}
	XtVaSetValues(tertris,
		XtNmenu, value + 1, NULL); /* Q */
}

static void
helpMenuCallback(Widget w, XtPointer clientData, XtPointer callData)
{
	int value = (int) ((size_t) clientData);
	Widget dialog;

#ifdef HAVE_ATHENA
	if (wmDeleteWindow == None)
		wmDeleteWindow = XInternAtom(XtDisplay(topLevel),
				"WM_DELETE_WINDOW", FALSE);
#endif
	switch (value) {
	case 0:
		dialog = descriptionDialog;
		break;
	case 1:
		dialog = featuresDialog;
		break;
	case 2:
		dialog = synopsisDialog;
		break;
	case 3:
		dialog = referencesDialog;
		break;
	case 4:
		dialog = aboutDialog;
		break;
	default:
		{
			char *buf;

			intCat(&buf, "helpMenuCallback: %d", value);
			XtWarning(buf);
			free(buf);
			return;
		}
	}
#ifdef HAVE_MOTIF
	XtManageChild(dialog);
#elif defined(HAVE_ATHENA)
	XtPopup(dialog, XtGrabNone);
	XSetWMProtocols(XtDisplay(topLevel),
		XtWindow(dialog), &wmDeleteWindow, 1);
#endif
}
#endif

#ifdef HAVE_MOTIF
static void
callbackGameRestart(Widget w, XtPointer clientData,
		XmAnyCallbackStruct * cbs)
{
	if (cbs->reason == XmCR_OK) {
		XtVaSetValues(tertris,
			XtNmenu, 4, NULL); /* menu choice */
	}
}

static void
closeDialog(Widget w, Widget shellDialog)
{
	XtDestroyWidget(shellDialog);
	setupDialog = NULL;
}

static Widget
createSetup(Widget w, char *title)
{
	Widget button, dialog, pane;
	Widget changerRowCol, unitsChangerRowCol;
	Widget fieldChangerRowCol, fillChangerRowCol;
	Widget toggleRowCol;
	int units, columns, fill;
	Boolean fixedFill, suddenAppear, diagonal, mixed;
	Boolean bonus, cw, plainPiece, showNext, grid, sound;
	char titleDsp[TITLE_FILE_LENGTH];
	XmString titleString = NULL;

	concatTitle(titleDsp, progDsp, ": ", title, TITLE_FILE_LENGTH);
	titleString = XmStringCreateSimple((char *) titleDsp);
	XtSetArg(args[0], XmNdeleteResponse, XmDESTROY);
	XtSetArg(args[1], XmNdialogTitle, titleString);
	dialog = XmCreateDialogShell(w, (char *) "Setup", args, 2);
	XmStringFree(titleString);
	pane = XtVaCreateWidget("pane", xmPanedWindowWidgetClass, dialog,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	changerRowCol = XtVaCreateManagedWidget("changerRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	toggleRowCol = XtVaCreateManagedWidget("toggleRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 10,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(tertris,
		XtNunits, &units,
		XtNcolumns, &columns,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNshowNext, &showNext,
		XtNgrid, &grid,
		XtNsound, &sound, NULL);
	unitsChangerRowCol = XtVaCreateManagedWidget("unitsChangerRowCol",
		xmRowColumnWidgetClass, changerRowCol,
		XmNnumColumns, 2,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_TIGHT, NULL);
#ifdef USE_SPIN
	createSpinner(unitsChangerRowCol, &unitsChanger,
		(char *) unitsLabel, units,
		MIN_UNITS, MAX_UNITS, 1,
		(XtCallbackProc) unitsChangeCallback);
#else
	createSlider(unitsChangerRowCol, &unitsChanger,
		(char *) unitsLabel, units,
		MIN_UNITS, MAX_UNITS, 1, SCROLL_SIZE,
		(XtCallbackProc) unitsChangeCallback);
#endif
	fieldChangerRowCol = XtVaCreateManagedWidget("changerFieldRowCol",
		xmRowColumnWidgetClass, changerRowCol,
		XmNnumColumns, 2,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_TIGHT, NULL);
#ifdef USE_SPIN
	createSpinner(fieldChangerRowCol, &fieldChanger,
		(char *) fieldLabel, columns,
		MIN_COLUMNS, MAX_COLUMNS, 2,
		(XtCallbackProc) fieldChangeCallback);
#else
	createSlider(fieldChangerRowCol, &fieldChanger,
		(char *) fieldLabel, columns,
		MIN_COLUMNS, MAX_COLUMNS, 2, SCROLL_SIZE,
		(XtCallbackProc) fieldChangeCallback);
#endif
	fillChangerRowCol = XtVaCreateManagedWidget("fillChangerRowCol",
		xmRowColumnWidgetClass, changerRowCol,
		XmNnumColumns, 2,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_TIGHT, NULL);
#ifdef USE_SPIN
	createSpinner(fillChangerRowCol, &fillChanger,
		(char *) fillLabel, fill,
		0, MAX_FILL, 2,
		(XtCallbackProc) fillChangeCallback);
#else
	createSlider(fillChangerRowCol, &fillChanger,
		(char *) fillLabel, fill,
		0, MAX_FILL, 2, SCROLL_SIZE,
		(XtCallbackProc) fillChangeCallback);
#endif
	fixedFillToggle = XtVaCreateManagedWidget("Fixed Fill",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(fixedFillToggle, XmNvalueChangedCallback,
		(XtCallbackProc) fixedFillCallback, (XtPointer) NULL);
	suddenAppearToggle = XtVaCreateManagedWidget("Sudden Appear",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(suddenAppearToggle, XmNvalueChangedCallback,
		(XtCallbackProc) suddenAppearCallback, (XtPointer) NULL);
	diagonalToggle = XtVaCreateManagedWidget("Diagonal",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(diagonalToggle, XmNvalueChangedCallback,
		(XtCallbackProc) diagonalCallback, (XtPointer) NULL);
	mixedToggle = XtVaCreateManagedWidget("Mixed",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(mixedToggle, XmNvalueChangedCallback,
		(XtCallbackProc) mixedCallback, (XtPointer) NULL);
	bonusToggle = XtVaCreateManagedWidget("Bonus",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(bonusToggle, XmNvalueChangedCallback,
		(XtCallbackProc) bonusCallback, (XtPointer) NULL);
	clockwiseToggle = XtVaCreateManagedWidget("Clockwise",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(clockwiseToggle, XmNvalueChangedCallback,
		(XtCallbackProc) clockwiseCallback, (XtPointer) NULL);
	plainPieceToggle = XtVaCreateManagedWidget("Plain Piece",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(plainPieceToggle, XmNvalueChangedCallback,
		(XtCallbackProc) plainPieceCallback, (XtPointer) NULL);
	showNextToggle = XtVaCreateManagedWidget("Show Next",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(showNextToggle, XmNvalueChangedCallback,
		(XtCallbackProc) showNextCallback, (XtPointer) NULL);
	gridToggle = XtVaCreateManagedWidget("Grid",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(gridToggle, XmNvalueChangedCallback,
		(XtCallbackProc) gridCallback, (XtPointer) NULL);
	soundToggle = XtVaCreateManagedWidget("Sound",
		xmToggleButtonWidgetClass, toggleRowCol, NULL);
	XtAddCallback(soundToggle, XmNvalueChangedCallback,
		(XtCallbackProc) soundCallback, (XtPointer) NULL);
	XmToggleButtonSetState(fixedFillToggle, fixedFill, True);
	XmToggleButtonSetState(suddenAppearToggle, suddenAppear, True);
	XmToggleButtonSetState(diagonalToggle, diagonal, True);
	XmToggleButtonSetState(mixedToggle, mixed, True);
	XmToggleButtonSetState(bonusToggle, bonus, True);
	XmToggleButtonSetState(clockwiseToggle, cw, True);
	XmToggleButtonSetState(plainPieceToggle, plainPiece, True);
	XmToggleButtonSetState(showNextToggle, showNext, True);
	XmToggleButtonSetState(gridToggle, grid, True);
	XmToggleButtonSetState(soundToggle, sound, True);
	button = XtVaCreateManagedWidget("OK",
		xmPushButtonGadgetClass, pane,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmALIGNMENT_CENTER,
		XmNshowAsDefault, True,
		XmNdefaultButtonShadowThickness, 1, NULL);
	XtAddCallback(button, XmNactivateCallback,
		(XtCallbackProc) closeDialog, dialog);
	XtManageChild(pane);
	return dialog;
}

#elif defined(HAVE_ATHENA)
static void
quitText(Widget w, XtPointer clientData, XtPointer callData)
{
	XtPopdown(setupDialogShell);
}
#endif

int
main(int argc, char **argv)
{
	int pixmapSize;
#if defined(HAVE_MOTIF) || defined(HAVE_ATHENA)
	Widget mainPanel, controlPanel;
	Widget levelRowCol, scoreRowCol;
#else
	char titleDsp[TITLE_FILE_LENGTH];
	Boolean showNext;
	int x, y, width;
#endif
#ifdef HAVE_MOTIF
	Widget pullDownMenu, widget;
	Widget menuBarPanel;
	Widget levelChangerRowCol;
	Widget messageRowCol;
	XmString fileString, playString;
	XmString quitString;
	XmString pauseString, clearString, setupString;
#ifdef EXTRA
	XmString incrementString, decrementString;
	XmString fixedFillString, suddenAppearString;
	XmString diagonalString, mixedString, bonusString;
	XmString cwString, plainPieceString, showNextString;
#endif
#elif defined(HAVE_ATHENA)
	XtActionsRec actions[] = {
		{(char *) "DeleteWindowProc", deleteWindowProc},
		{(char *) "ClosePopupPanel", (XtActionProc) closePopupPanel}
	};
	String fallbackResources[] = {
		(char *) "?.translations: #override <Message>WM_PROTOCOLS: DeleteWindowProc()",
		(char *) "?.TransientShell.translations: #override <Message>WM_PROTOCOLS: ClosePopupPanel()",
		NULL
	};
	Widget levelBox, unitsBox, fieldBox, fillBox;
	Widget fixedFillBox, suddenAppearBox, diagonalBox, mixedBox, bonusBox;
	Widget clockwiseBox, plainPieceBox, showNextBox, gridBox, soundBox;
	Widget rowsLabel, scoreLabel, maxScoreLabel;
	Widget rowsRowCol, maxScoreRowCol;
	Widget menuBar, panedPanel;
	Widget w, helpMenu, helpLabel;
	Widget setupForm, setupQuitButton;
	int level, units, columns, fill;
	Boolean fixedFill, suddenAppear, diagonal, mixed, bonus;
	Boolean cw, plainPiece, showNext, grid, sound;
	unsigned int i;
#endif

	progDsp = strrchr(argv[0], '/');
	if (progDsp != NULL)
		progDsp++;
	if (progDsp == NULL)
		progDsp = argv[0];
	topLevel = XtVaAppInitialize(NULL, "Tertris",
		options, XtNumber(options), &argc, argv,
#ifdef HAVE_ATHENA
		fallbackResources,
#else
		NULL,
#endif
		NULL);
	appCon = XtWidgetToApplicationContext(topLevel);
	if (argc != 1)
		usage(argv[0], synopsisHelp);
#ifdef HAVE_MOTIF
	menuBarPanel = XtVaCreateManagedWidget("menuBarPanel",
		xmPanedWindowWidgetClass, topLevel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	fileString = XmStringCreateSimple((char *) "File");
	playString = XmStringCreateSimple((char *) "Play");
	menuBar = XmVaCreateSimpleMenuBar(menuBarPanel, (char *) "menuBar",
		XmVaCASCADEBUTTON, fileString, 'F',
		XmVaCASCADEBUTTON, playString, 'P', NULL);
	XmStringFree(fileString);
	XmStringFree(playString);
	quitString = XmStringCreateSimple((char *) "Exit");
	(void) XmVaCreateSimplePulldownMenu(menuBar, (char *) "fileMenu",
		0, fileMenuCallback,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quitString, 'x', NULL, NULL, NULL);
	XmStringFree(quitString);
	pauseString = XmStringCreateSimple((char *) "Pause");
	clearString = XmStringCreateSimple((char *) "Clear");
	setupString = XmStringCreateSimple((char *) "Setup");
#ifdef EXTRA
	incrementString = XmStringCreateSimple((char *) "Increment");
	decrementString = XmStringCreateSimple((char *) "Decrement");
	fixedFillString = XmStringCreateSimple((char *) "Fixed Fill");
	suddenAppearString = XmStringCreateSimple((char *) "Sudden Appear");
	diagonalString = XmStringCreateSimple((char *) "Diagonal");
	mixedString = XmStringCreateSimple((char *) "Mixed");
	bonusString = XmStringCreateSimple((char *) "Bonus");
	cwString = XmStringCreateSimple((char *) "Clockwise");
	plainPieceString = XmStringCreateSimple((char *) "Plain Piece");
	showNextString = XmStringCreateSimple((char *) "Show Next");
#endif
	(void) XmVaCreateSimplePulldownMenu(menuBar, (char *) "playMenu",
		1, playMenuCallback,
		XmVaPUSHBUTTON, pauseString, 'P', NULL, NULL,
		XmVaPUSHBUTTON, clearString, 'C', NULL, NULL,
		XmVaPUSHBUTTON, setupString, 'S', NULL, NULL,
#ifdef EXTRA
		XmVaPUSHBUTTON, incrementString, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrementString, 'D', NULL, NULL,
		XmVaPUSHBUTTON, fixedFillString, 'x', NULL, NULL,
		XmVaPUSHBUTTON, suddenAppearString, 'r', NULL, NULL,
		XmVaPUSHBUTTON, diagonalString, 'g', NULL, NULL,
		XmVaPUSHBUTTON, mixedString, 'M', NULL, NULL,
		XmVaPUSHBUTTON, bonusString, 'B', NULL, NULL,
		XmVaPUSHBUTTON, cwString, 'w', NULL, NULL,
		XmVaPUSHBUTTON, plainPieceString, 'P', NULL, NULL,
		XmVaPUSHBUTTON, showNextString, 'N', NULL, NULL,
#endif
		NULL);
	XmStringFree(pauseString);
	XmStringFree(clearString);
	XmStringFree(setupString);
#ifdef EXTRA
	XmStringFree(incrementString);
	XmStringFree(decrementString);
	XmStringFree(fixedFillString);
	XmStringFree(suddenAppearString);
	XmStringFree(diagonalString);
	XmStringFree(mixedString);
	XmStringFree(bonusString);
	XmStringFree(cwString);
	XmStringFree(plainPieceString);
	XmStringFree(showNextString);
#endif
	pullDownMenu = XmCreatePulldownMenu(menuBar,
		(char *) "helpPullDown", NULL, 0);
	widget = XtVaCreateManagedWidget("Help",
		xmCascadeButtonWidgetClass, menuBar,
		XmNsubMenuId, pullDownMenu,
		XmNmnemonic, 'H', NULL); /* mnemonic warning on Cygwin */
	XtVaSetValues(menuBar,
		XmNmenuHelpWidget, widget, NULL);
	widget = XtVaCreateManagedWidget("Description...",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'D', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuCallback, (char *) 0);
	widget = XtVaCreateManagedWidget("Features...",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'F', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuCallback, (char *) 1);
	widget = XtVaCreateManagedWidget("Synopsis...",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'S', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuCallback, (char *) 2);
	widget = XtVaCreateManagedWidget("References...",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'R', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuCallback, (char *) 3);
	widget = XtVaCreateManagedWidget("About...",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'A', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuCallback, (char *) 4);
	XtManageChild(menuBar);
	restartDialog = createQuery(topLevel, (char *) "Reset Query",
		(char *) "Are you sure you want to restart?",
		(XtCallbackProc) callbackGameRestart);
	mainPanel = XtVaCreateManagedWidget("mainPanel",
		xmPanedWindowWidgetClass, menuBarPanel,
		XmNorientation, XmHORIZONTAL, NULL);
	controlPanel = XtVaCreateManagedWidget("controlPanel",
		xmPanedWindowWidgetClass, mainPanel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	levelRowCol = XtVaCreateManagedWidget("levelRowCol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 1,
		XmNorientation, XmHORIZONTAL, NULL);
	scoreRowCol = XtVaCreateManagedWidget("Rowcol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	levelChangerRowCol = XtVaCreateManagedWidget("levelChangerRowCol",
		xmRowColumnWidgetClass, levelRowCol,
		XmNnumColumns, 2,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_TIGHT, NULL);
#ifdef USE_SPIN
	createSpinner(levelChangerRowCol, &levelChanger,
		(char *) levelLabel, MIN_LEVELS,
		MIN_LEVELS, MAX_LEVELS, 2,
		(XtCallbackProc) levelChangeCallback);
#else
	createSlider(levelChangerRowCol, &levelChanger,
		(char *) levelLabel, MIN_LEVELS,
		MIN_LEVELS, MAX_LEVELS, 2, SCROLL_SIZE,
		(XtCallbackProc) levelChangeCallback);
#endif
	(void) XtVaCreateManagedWidget("rowsText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Rows", 5, NULL);
	rowsText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("scoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Score", 6, NULL);
	scoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("maxScoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Max Score", 10, NULL);
	maxScoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	nextTertris = XtVaCreateManagedWidget("tertrisNext",
		nextTertrisWidgetClass, controlPanel, NULL);
	messageRowCol = XtVaCreateManagedWidget("messageRowCol",
		xmRowColumnWidgetClass, controlPanel, NULL);
	message = XtVaCreateManagedWidget("Play Tertris!",
		xmLabelWidgetClass, messageRowCol, NULL);
	tertris = XtVaCreateManagedWidget("tertris",
		tertrisWidgetClass, mainPanel, NULL);
#elif defined(HAVE_ATHENA)
	XtAppAddActions(appCon, actions, XtNumber(actions));
	createHelp(topLevel, &descriptionDialog, (char *) "Description",
		(char *) descriptionHelp, (XtCallbackProc) closePopupPanel2);
	createScrollHelp(topLevel, &featuresDialog, (char *) "Features",
		(char *) featuresHelp, (XtCallbackProc) closePopupPanel2);
	createHelp(topLevel, &synopsisDialog, (char *) "Synopsis",
		(char *) synopsisHelp, (XtCallbackProc) closePopupPanel2);
	createHelp(topLevel, &referencesDialog, (char *) "References",
		(char *) referencesHelp, (XtCallbackProc) closePopupPanel2);
	createHelp(topLevel, &aboutDialog, (char *) "About",
		(char *) aboutHelp, (XtCallbackProc) closePopupPanel2);
	mainPanel = XtVaCreateManagedWidget("mainPanel",
		panedWidgetClass, topLevel, NULL);
	menuBar = XtVaCreateManagedWidget("MenuBar",
		formWidgetClass, mainPanel,
		XtNborderWidth, 1, NULL);
	/* multiple pulldowns were causing an error */
	helpLabel = XtVaCreateManagedWidget("Menu",
		menuButtonWidgetClass, menuBar,
		XtNborderWidth, 0, NULL);
	helpMenu = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, helpLabel, NULL);
	for (i = 0; i < numPlayTypes; i++) {
		w = XtVaCreateManagedWidget(playTypes[i],
			smeBSBObjectClass, helpMenu, NULL);
		XtAddCallback(w,
			XtNcallback, (XtCallbackProc) playMenuCallback,
			(XtPointer) (size_t) i);
	}
	XtVaCreateManagedWidget("menuLine",
		smeLineObjectClass, helpMenu, NULL);
	for (i = 0; i < numHelpTypes; i++) {
		w = XtVaCreateManagedWidget(helpTypes[i],
			smeBSBObjectClass, helpMenu, NULL);
		XtAddCallback(w,
			XtNcallback, (XtCallbackProc) helpMenuCallback,
			(XtPointer) (size_t) i);
	}
	XtVaCreateManagedWidget("menuLine",
		smeLineObjectClass, helpMenu, NULL);
	for (i = 0; i < numFileTypes; i++) {
		w = XtVaCreateManagedWidget(fileTypes[i],
			smeBSBObjectClass, helpMenu, NULL);
		XtAddCallback(w,
			XtNcallback, (XtCallbackProc) fileMenuCallback,
			(XtPointer) (size_t) i);
	}
	setupDialogShell = XtCreatePopupShell("Setup",
		transientShellWidgetClass, topLevel, NULL, 0);
	setupForm = XtVaCreateManagedWidget("form",
		formWidgetClass, setupDialogShell, NULL);
	unitsBox = XtVaCreateManagedWidget("unitsBox",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0, NULL);
	fieldBox = XtVaCreateManagedWidget("fieldBox",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, unitsBox, NULL);
	fillBox = XtVaCreateManagedWidget("fillBox",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, fieldBox, NULL);
	fixedFillBox = XtVaCreateManagedWidget("Fixed Fill",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, fillBox, NULL);
	suddenAppearBox = XtVaCreateManagedWidget("Sudden Appear",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, fixedFillBox, NULL);
	diagonalBox = XtVaCreateManagedWidget("Diagonal",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, suddenAppearBox, NULL);
	mixedBox = XtVaCreateManagedWidget("Mixed",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, diagonalBox, NULL);
	bonusBox = XtVaCreateManagedWidget("Bonus",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, mixedBox, NULL);
	clockwiseBox = XtVaCreateManagedWidget("Clockwise",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, bonusBox, NULL);
	plainPieceBox = XtVaCreateManagedWidget("Plain Piece",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, clockwiseBox, NULL);
	showNextBox = XtVaCreateManagedWidget("Show Next",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, plainPieceBox, NULL);
	gridBox = XtVaCreateManagedWidget("Grid",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, showNextBox, NULL);
	soundBox = XtVaCreateManagedWidget("Sound",
		boxWidgetClass, setupForm,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNfromVert, gridBox, NULL);
	setupQuitButton = XtVaCreateManagedWidget("Quit",
		commandWidgetClass, setupForm,
		XtNfromVert, soundBox, NULL);
	XtAddCallback(setupQuitButton,
		XtNcallback, quitText, (XtPointer) setupQuitButton);
	panedPanel = XtVaCreateManagedWidget("panedPanel",
		formWidgetClass, mainPanel,
		XtNborderWidth, 0,
		XtNfromHoriz, menuBar, NULL);
	controlPanel = XtVaCreateManagedWidget("controlPanel",
		formWidgetClass, panedPanel,
		XtNborderWidth, 0, NULL);
	levelRowCol = XtVaCreateManagedWidget("levelRowCol",
		formWidgetClass, controlPanel,
		XtNborderWidth, 0, NULL);
	levelBox = XtVaCreateManagedWidget("levelBox",
		boxWidgetClass, levelRowCol,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0, NULL);
	rowsRowCol = XtVaCreateManagedWidget("rowsRowCol",
		formWidgetClass, controlPanel,
		XtNborderWidth, 0,
		XtNfromVert, levelRowCol, NULL);
	rowsLabel = XtVaCreateManagedWidget("Rows     ",
		labelWidgetClass, rowsRowCol,
		XtNborderWidth, 0, NULL);
	rowsText = XtVaCreateManagedWidget("0",
		labelWidgetClass, rowsRowCol,
		XtNwidth, 32,
		XtNfromHoriz, rowsLabel, NULL);
	scoreRowCol = XtVaCreateManagedWidget("scoreRowCol",
		formWidgetClass, controlPanel,
		XtNborderWidth, 0,
		XtNfromVert, rowsRowCol, NULL);
	scoreLabel = XtVaCreateManagedWidget("Score    ",
		labelWidgetClass, scoreRowCol,
		XtNborderWidth, 0, NULL);
	scoreText = XtVaCreateManagedWidget("0",
		labelWidgetClass, scoreRowCol,
		XtNwidth, 64,
		XtNfromHoriz, scoreLabel, NULL);
	maxScoreRowCol = XtVaCreateManagedWidget("maxScoreRowCol",
		formWidgetClass, controlPanel,
		XtNborderWidth, 0,
		XtNfromVert, scoreRowCol, NULL);
	maxScoreLabel = XtVaCreateManagedWidget("Max Score",
		labelWidgetClass, maxScoreRowCol,
		XtNborderWidth, 0, NULL);
	maxScoreText = XtVaCreateManagedWidget("0",
		labelWidgetClass, maxScoreRowCol,
		XtNwidth, 64,
		XtNfromHoriz, maxScoreLabel, NULL);
	nextTertris = XtVaCreateManagedWidget("tertrisNext",
		nextTertrisWidgetClass, controlPanel,
		XtNfromVert, maxScoreRowCol, NULL);
	tertris = XtVaCreateManagedWidget("tertris",
		tertrisWidgetClass, panedPanel,
		XtNfromHoriz, controlPanel, NULL);
	XtVaGetValues(tertris,
		XtNlevel, &level,
		XtNunits, &units,
		XtNcolumns, &columns,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNshowNext, &showNext,
		XtNgrid, &grid,
		XtNsound, &sound, NULL);
#ifdef USE_SPIN
	createSpinner(levelBox, &levelSliderLabel,
		levelLabel, 10, level, MIN_LEVELS, MAX_LEVELS, True,
		levelUpCallback, levelDownCallback);
	createSpinner(unitsBox, &unitsSliderLabel,
		unitsLabel, 12, units, MIN_UNITS, MAX_UNITS, True,
		unitsUpCallback, unitsDownCallback);
	createSpinner(fieldBox, &fieldSliderLabel,
		fieldLabel, 12, columns, MIN_COLUMNS, MAX_COLUMNS, True,
		fieldUpCallback, fieldDownCallback);
	createSpinner(fillBox, &fillSliderLabel,
		fillLabel, 12, fill, 0, MAX_FILL, True,
		fillUpCallback, fillDownCallback);
#else
	createSlider(levelBox, &levelSliderLabel, &levelSlider,
		levelLabel, 0, level, MIN_LEVELS, MAX_LEVELS, True,
		SCROLL_SIZE, levelChangeCallback, levelChangeCallback);
	createSlider(unitsBox, &unitsSliderLabel, &unitsSlider,
		unitsLabel, 12, units, MIN_UNITS, MAX_UNITS, True,
		SCROLL_SIZE, unitsChangeCallback, unitsChangeCallback);
	createSlider(fieldBox, &fieldSliderLabel, &fieldSlider,
		fieldLabel, 12, columns, MIN_COLUMNS, MAX_COLUMNS, True,
		SCROLL_SIZE, fieldChangeCallback, fieldChangeCallback);
	createSlider(fillBox, &fillSliderLabel, &fillSlider,
		fillLabel, 12, fill, 0, MAX_FILL, True,
		SCROLL_SIZE, fillChangeCallback, fillChangeCallback);
#endif
	createToggle(fixedFillBox, &fixedFillToggle,
		"Fixed Fill:", 15,
		fixedFill, (XtCallbackProc) fixedFillCallback);
	createToggle(suddenAppearBox, &suddenAppearToggle,
		"Sudden Appear:", 15,
		suddenAppear, (XtCallbackProc) suddenAppearCallback);
	createToggle(diagonalBox, &diagonalToggle,
		"Diagonal:", 15,
		diagonal, (XtCallbackProc) diagonalCallback);
	createToggle(mixedBox, &mixedToggle,
		"Mixed:", 15,
		mixed, (XtCallbackProc) mixedCallback);
	createToggle(bonusBox, &bonusToggle,
		"Bonus:", 15,
		bonus, (XtCallbackProc) bonusCallback);
	createToggle(clockwiseBox, &clockwiseToggle,
		"Clockwise:", 15,
		cw, (XtCallbackProc) clockwiseCallback);
	createToggle(plainPieceBox, &plainPieceToggle,
		"Plain Piece:", 15,
		plainPiece, (XtCallbackProc) plainPieceCallback);
	createToggle(showNextBox, &showNextToggle,
		"Show Next:", 15,
		showNext, (XtCallbackProc) showNextCallback);
	createToggle(gridBox, &gridToggle,
		"Grid:", 15,
		grid, (XtCallbackProc) gridCallback);
	createToggle(soundBox, &soundToggle,
		"Sound:", 15,
		sound, (XtCallbackProc) soundCallback);
	XtRealizeWidget(topLevel);
	if (wmDeleteWindow == None)
		wmDeleteWindow = XInternAtom(XtDisplay(topLevel),
			"WM_DELETE_WINDOW", FALSE);
	XSetWMProtocols(XtDisplay(topLevel), XtWindow(topLevel),
		&wmDeleteWindow, 1);
#else
	tertris = XtVaCreateManagedWidget("tertris",
		tertrisWidgetClass, topLevel, NULL);
	concatTitle(titleDsp, progDsp, ": ", "Next", TITLE_FILE_LENGTH);
	shell = XtAppCreateShell(progDsp, titleDsp,
		topLevelShellWidgetClass, XtDisplay(topLevel), NULL, 0);
	nextTertris = XtVaCreateManagedWidget("nextTertris",
		nextTertrisWidgetClass, shell, NULL);
#endif
	XtVaGetValues(tertris,
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
		XtNshowNext, &showNext,
#endif
		XtNpixmapSize, &pixmapSize, NULL);
#ifdef HAVE_XPM
	{
		XpmAttributes xpmAttributes;
		XpmColorSymbol transparentColor[1] = {{NULL,
			(char *) "none", 0 }};
		Pixel bg;

		xpmAttributes.valuemask = XpmColorSymbols | XpmCloseness;
		xpmAttributes.colorsymbols = transparentColor;
		xpmAttributes.numsymbols = 1;
		xpmAttributes.closeness = 40000;
		XtVaGetValues(topLevel,
			XtNbackground, &bg, NULL);
		transparentColor[0].pixel = bg;
		(void) XpmCreatePixmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			RESIZE_XPM(pixmapSize), &pixmap, NULL,
			&xpmAttributes);
	}
	if (pixmap == (Pixmap) NULL)
#endif
		pixmap = XCreateBitmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			DEFINE_XBM);
	XtVaSetValues(topLevel,
#ifdef HAVE_MOTIF
		XmNkeyboardFocusPolicy, XmPOINTER, /* not XmEXPLICIT */
#else
		XtNinput, True,
#endif
		XtNiconPixmap, pixmap, NULL);
	XtAddCallback(tertris, XtNselectCallback,
		(XtCallbackProc) callbackGame, (XtPointer) NULL);
	XtAddCallback(nextTertris, XtNselectCallback,
		(XtCallbackProc) callbackNextGame, (XtPointer) NULL);
#ifdef HAVE_MOTIF
	descriptionDialog = createHelp(menuBar, (char *) "Description",
		(char *) descriptionHelp);
	featuresDialog = createScrollHelp(menuBar, (char *) "Features",
		(char *) featuresHelp, pixmap);
	synopsisDialog = createHelp(menuBar, (char *) "Synopsis",
		(char *) synopsisHelp);
	referencesDialog = createHelp(menuBar, (char *) "References",
		(char *) referencesHelp);
	aboutDialog = createHelp(menuBar, (char *) "About",
		(char *) aboutHelp);
#endif
	initialize(tertris);
	XtRealizeWidget(topLevel);
	XGrabButton(XtDisplay(tertris), (unsigned int) AnyButton,
		AnyModifier, XtWindow(tertris), TRUE,
		(unsigned int) (ButtonPressMask),
		GrabModeAsync, GrabModeAsync, XtWindow(tertris),
		XCreateFontCursor(XtDisplay(tertris), XC_crosshair));
#if !defined(HAVE_MOTIF) && !defined(HAVE_ATHENA)
	XtVaGetValues(topLevel,
		XtNx, &x,
		XtNy, &y,
		XtNwidth, &width, NULL);
	XtVaSetValues(shell,
		XtNx, x + width,
		XtNy, y,
		XtNiconPixmap, pixmap,
		XtNinput, True, NULL);
	if (showNext)
		XtRealizeWidget(shell);
#endif
	(void) XtAppAddTimeOut(appCon, (unsigned long) 1000,
		playingTimer, (XtPointer) tertris);
	XtAppMainLoop(appCon);

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
#endif
